@@ -33,6 +33,11 @@
 ;;;; The functions RUN!, !, !! and !!! are convenient wrappers around
 ;;;; RUN and EXPLAIN.
 
+(deftype on-problem-action ()
+  '(member :debug :backtrace nil))
+
+(declaim (type on-problem-action *on-error* *on-failure*))
+
 (defvar *on-error* nil
   "The action to perform on error:
 - :DEBUG if we should drop into the debugger
@@ -56,6 +61,12 @@ OBSOLETE: superseded by *ON-FAILURE*")
 (defparameter *print-names* t
   "T if we should print test running progress, NIL otherwise.")
 
+(defparameter *test-dribble-indent* (make-array 0
+                                        :element-type 'character
+                                        :fill-pointer 0
+                                        :adjustable t)
+  "Used to indent tests and test suites in their parent suite")
+
 (defun import-testing-symbols (package-designator)
   (import '(5am::is 5am::is-true 5am::is-false 5am::signals 5am::finishes)
           package-designator))
@@ -128,18 +139,23 @@ run."))
           (not (satisfies-depends-p #'notany))
           (:before (every #'(lambda (dep)
                               (let ((status (status (get-test dep))))
-                                (eql :unknown status)))
+                                (if (eql :unknown status)
+                                    (run-resolving-dependencies (get-test dep))
+                                    status)))
                           (cdr depends-on)))))))
 
 (defun results-status (result-list)
   "Given a list of test results (generated while running a test)
-  return true if all of the results are of type TEST-PASSED,
-  fail otherwise.
-  Returns a second value, which is the set of failed tests."
+  return true if no results are of type TEST-FAILURE.  Returns second
+  and third values, which are the set of failed tests and skipped
+  tests respectively."
   (let ((failed-tests
-          (remove-if #'test-passed-p result-list)))
+          (remove-if-not #'test-failure-p result-list))
+        (skipped-tests
+          (remove-if-not #'test-skipped-p result-list)))
     (values (null failed-tests)
-            failed-tests)))
+            failed-tests
+            skipped-tests)))
 
 (defun return-result-list (test-lambda)
   "Run the test function TEST-LAMBDA and return a list of all
@@ -186,6 +202,8 @@ run."))
                          (handler-case
                              (let ((*readtable* (copy-readtable))
                                    (*package* (runtime-package test)))
+                               (when *print-names*
+                                   (format *test-dribble* "~%~ARunning test ~A " *test-dribble-indent* (name test)))
                                (if (collect-profiling-info test)
                                    ;; Timing info doesn't get collected ATM, we need a portable library
                                    ;; (setf (profiling-info test) (collect-timing (test-lambda test)))
@@ -216,8 +234,6 @@ run."))
   !!, !!!"))
 
 (defmethod %run ((test test-case))
-  (when *print-names*
-    (format *test-dribble* "~% Running test ~A " (name test)))
   (run-resolving-dependencies test))
 
 (defmethod %run ((tests list))
@@ -225,12 +241,13 @@ run."))
 
 (defmethod %run ((suite test-suite))
   (when *print-names*
-    (format *test-dribble* "~%Running test suite ~A" (name suite)))
+    (format *test-dribble* "~%~ARunning test suite ~A" *test-dribble-indent* (name suite)))
   (let ((suite-results '()))
     (flet ((run-tests ()
              (loop
                 for test being the hash-values of (tests suite)
                 do (%run test))))
+      (vector-push-extend #\space *test-dribble-indent*)
       (unwind-protect
            (bind-run-state ((result-list '()))
              (unwind-protect
@@ -241,6 +258,7 @@ run."))
                       (run-tests)))
              (setf suite-results result-list
                    (status suite) (every #'test-passed-p suite-results)))
+        (vector-pop *test-dribble-indent*)
         (with-run-state (result-list)
           (setf result-list (nconc result-list suite-results)))))))
 
@@ -292,13 +310,13 @@ performed by the !, !! and !!! functions."
   (let ((*on-error*
           (or *on-error* (cond
                            (*debug-on-error*
-                            (format *test-dribble* "*DEBUG-ON-ERROR* is obsolete. Use *ON-ERROR.")
+                            (format *test-dribble* "*DEBUG-ON-ERROR* is obsolete. Use *ON-ERROR*.")
                             :debug)
                            (t nil))))
         (*on-failure*
           (or *on-failure* (cond
                            (*debug-on-failure*
-                            (format *test-dribble* "*DEBUG-ON-FAILURE* is obsolete. Use *ON-FAILURE.")
+                            (format *test-dribble* "*DEBUG-ON-FAILURE* is obsolete. Use *ON-FAILURE*.")
                             :debug)
                            (t nil)))))
     (funcall *!*)))
@@ -320,7 +338,7 @@ performed by the !, !! and !!! functions."
 SUMMARY can be :END to print a summary at the end, :SUITE to print it
 after each suite or NIL to skip explanations."
   (check-type summary (member nil :suite :end))
-  (loop :for suite :in (cons 'NIL (sort (copy-list *toplevel-suites*) #'string<=))
+  (loop :for suite :in (cons 'nil (sort (copy-list *toplevel-suites*) #'string<=))
         :for results := (if (suite-emptyp suite) nil (run suite))
         :when (consp results)
           :collect results :into all-results
